#
# Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
# Use is subject to license terms.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free  Software Foundation; either version 2 of the License, or
# (at your option)  any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA.
#

from string import ascii_letters
import virtconv.vmcfg as vmcfg
import virtconv.diskcfg as diskcfg
import virtconv.netdevcfg as netdevcfg
import virtconv.formats as formats
import virtinst.FullVirtGuest as fv

import os
import stat
import re

bootdevs = {
    vmcfg.VM_TYPE_PV: "",
    vmcfg.VM_TYPE_HVM: "<boot dev='hd' />",
}

consoles = {
    vmcfg.VM_TYPE_PV: "",
    vmcfg.VM_TYPE_HVM: "<console type='pty' />",
}

disk_template = """
<disk %(typeattr)s device='%(device)s'>
 %(hostdev)s
 <target dev='%(prefix)s%(dev)s' />
 %(readonly)s
</disk>
"""

netdev_template = """
<interface type='%(type)s'>
 %(source)s
 %(mac)s
 %(model)s
</interface>
"""

instance_template = """
<domain type='xen'>
  <name>%(name)s</name>
  %(bootloader)s
  <os>
    <type>%(type)s</type>
    %(loader)s
    %(bootdev)s
  </os>
  <memory>%(memory)s</memory>
  <vcpu>%(nr_vcpus)s</vcpu>
  <on_poweroff>destroy</on_poweroff>
  <on_reboot>restart</on_reboot>
  <on_crash>restart</on_crash>
  <clock offset='%(clock)s'/>
  <features>
    %(acpi)s
    %(apic)s
    <pae />
  </features>
  <distro>
    <type>%(distrotype)s</type>
    <variant>%(distrovariant)s</variant>
  </distro>
  <devices>
    %(emulator)s
    %(disks)s
    %(netdevs)s
    <graphics type='vnc' port='-1' />
    <input type='mouse' bus='ps2' />
    %(usbtablet)s
    %(console)s
  </devices>
</domain>
"""

def export_netdevs(vm):
    """
    Export code for the network devices.
    """

    netdevs = []

    for netdev in vm.netdevs.values():
        mac = ""
        if netdev.mac != "auto":
            mac = "<mac address='%s' />" % netdev.mac

        nettype = {
            netdevcfg.NETDEV_TYPE_DEV : "ethernet",
            netdevcfg.NETDEV_TYPE_BRIDGE : "bridge",
            netdevcfg.NETDEV_TYPE_UNKNOWN : "bridge",
            netdevcfg.NETDEV_TYPE_NETWORK : "network",
        }.get(netdev.type)

        source = ""
        if netdev.source:
            srcattr = nettype
            if netdev.type == netdevcfg.NETDEV_TYPE_DEV:
                srcattr = "dev"
            source = "<source %s='%s' />" % (srcattr, netdev.source)

        # FIXME: should warn here
        if not nettype:
            continue

        model = ""
        if netdev.driver:
            model = "<model type='%s' />" % netdev.driver

        netdevs.append(netdev_template % {
            "type" : nettype,
            "source" : source,
            "mac" : mac,
            "model": model,
        })

    return netdevs

def export_disks(vm):
    """
    Export code for the disks.  Slightly tricky for two reasons.
    
    We can't handle duplicate disks: some vmx files define SCSI/IDE devices
    that point to the same storage, and Xen isn't happy about that. We
    just ignore any entries that have duplicate paths.

    Since there is no meaningful SCSI support in rombios/qemu, we
    forcibly switch the disks to IDE, and expect the guest OS to cope
    (which at least Linux does admirably).
    """

    out = []
    paths = []

    disks = {}

    for (bus, instance), disk in sorted(vm.disks.iteritems()):

        if disk.path and disk.path in paths:
            continue

        if bus == "scsi":
            instance = 0
            while disks.get(("ide", instance)):
                instance += 1

        disks[("ide", instance)] = disk

        if disk.path:
            paths += [ disk.path ]

    for (bus, instance), disk in sorted(disks.iteritems()):

        if not disk.path:
            typeattr = ""
            hostdev = ""

            # It's quite common for .vmx files to reference a
            # non-existent ISO (which was cleaned up in vmx_parser).
            # Just skip them.
            if disk.type == diskcfg.DISK_TYPE_ISO:
                continue
        else:
            # Of course, this file path might be relative, so we won't be
            # able to stat() it.  In such a case, it's almost certainly a
            # file anyway, so the fallback is fine.
            typeattr = "type='file'"
            hostdev = ("<driver name='file' />\n"
                "<source file='%s' />\n" % disk.path)

            try:
                if stat.S_ISBLK(os.stat(disk.path)[0]):
                    typeattr = "type='block'"
                    hostdev = ("<driver name='phy' />\n"
                        "<source dev='%s' />\n" % disk.path)
            except:
                pass

            if disk.format == diskcfg.DISK_FORMAT_VDISK:
                hostdev = ("<driver name='tap' type='vdisk' />\n"
                    "<source file='%s' />\n" % disk.path)

        device = "disk"
        readonly = ""

        if (disk.type == diskcfg.DISK_TYPE_CDROM or
            disk.type == diskcfg.DISK_TYPE_ISO):
            device = "cdrom"
            readonly = "<readonly />"

        bus = "ide"

        disk_prefix = "xvd"
        if vm.type == vmcfg.VM_TYPE_HVM:
            if bus == "ide":
                disk_prefix = "hd"
            else:
                disk_prefix = "sd"

        # FIXME: needs updating for later Xen enhancements; need to
        # implement capabilities checking for max disks etc.
        drive_nr = ascii_letters[int(instance) % 26]

        instance += 1

        out.append(disk_template % {
            "typeattr" : typeattr,
            "device" : device,
            "hostdev" : hostdev,
            "prefix" : disk_prefix,
            "dev" :  drive_nr,
            "readonly" : readonly,
        })
 
    return out

def export_os_params(vm):
    """
    Export OS-specific parameters.
    """
    ostype = None
    osvariant = None

    ostype = fv._OS_TYPES.get(vm.os_type)
    if ostype:
        osvariant = ostype.get('variants').get(vm.os_variant)

    def get_os_val(key, default):
        val = None
        if osvariant:
            val = osvariant.get(key)
        if not val and ostype:
            val = ostype.get(key)
        if not val:
            val = default
        return val

    acpi = ""
    if vm.noacpi is False and get_os_val("acpi", True):
        acpi = "<acpi />"

    apic = ""
    if vm.noapic is False and get_os_val("apic", False):
        apic = "<apic />"

    clock = get_os_val("clock", "utc")

    usbtablet = ""
    if get_os_val("input", [ "tablet", "usb" ])[0] == "tablet":
        usbtablet = "<input type='tablet' bus='usb' />"

    return acpi, apic, clock, usbtablet


class virtinstance_parser(formats.parser):
    """
    Support for libvirt's domain instance format as defined (mostly) by
    libvirt.rng.  Currently, this only produces domains for Xen.

    This is a somewhat challenging format as it encodes significant
    amounts of information specific to a particular platform (for
    example, the pygrub path.  For now, we'll assume it's for the current
    platform.  Later, we'd take an optional hypervisor connection.

    Known limitations:

    - Only handles bridge, ethernet, and network type netdevs
    - doesn't handle netdev script or IP address
    """

    name = "virt-instance"
    suffix = ".virt-instance.xml"
    can_import = False
    can_export = True
    can_identify = False

    @staticmethod
    def identify_file(input_file):
        """
        Return True if the given file is of this format.
        """
        raise NotImplementedError

    @staticmethod
    def import_file(input_file):
        """
        Import a configuration file.  Raises if the file couldn't be
        opened, or parsing otherwise failed.
        """
        raise NotImplementedError

    @staticmethod
    def export_file(vm, output_file):
        """
        Export a configuration file.
        @vm vm configuration instance
        @file Output file

        Raises ValueError if configuration is not suitable, or another
        exception on failure to write the output file.
        """

        if not vm.memory:
            raise ValueError("VM must have a memory setting")

        # xend wants the name to match r'^[A-Za-z0-9_\-\.\:\/\+]+$'
        # The schema wants r'^[A-Za-z0-9_\-\.\:\/\+&amp;]+$'
        vmname = re.sub(r'[^A-Za-z0-9_\-\.\:\/\+]+',  '_', vm.name)

        vmtype = "xen"
        if vm.type == vmcfg.VM_TYPE_HVM:
            vmtype = "hvm"

        bootloader = ""
        loader = "/usr/lib/xen/boot/hvmloader"
        emulator = "<emulator>/usr/lib/xen/bin/qemu-dm</emulator>"

        if vm.type == vmcfg.VM_TYPE_PV:
            bootloader = loader
            emulator = ""
            loader = ""

        if bootloader:
            bootloader = "<bootloader>%s</bootloader>" % bootloader
        if loader:
            loader = "<loader>%s</loader>" % loader

        netdevs = export_netdevs(vm)
        disks = export_disks(vm)
  
        acpi, apic, clock, usbtablet = export_os_params(vm)

        dtype = vm.os_type
        if not dtype:
            dtype = "unknown"
        dvariant = vm.os_variant
        if not dvariant:
            dvariant = "unknown"

        out = instance_template % {
            "name" : vmname,
            "bootloader" : bootloader,
            "type" : vmtype,
            "loader"  : loader,
            "bootdev" : bootdevs[vm.type],
            # Mb to Kb
            "memory" : int(vm.memory) * 1024,
            "nr_vcpus" : vm.nr_vcpus,
            "clock" : clock,
            "acpi" : acpi,
            "apic" : apic,
            "emulator" : emulator,
            "disks" : "".join(disks),
            "netdevs" : "".join(netdevs),
            "usbtablet" : usbtablet,
            "console" : consoles[vm.type],
            "distrotype" : dtype,
            "distrovariant" : dvariant
        }

        outfile = open(output_file, "w")
        outfile.writelines(out)
        outfile.close()

formats.register_parser(virtinstance_parser)
